@@ -3,20 +3,19 @@
 """
 # PS: the coincidential module name is intentional ;)
 
-import json
 from logging import getLogger
 from pathlib import Path
-from typing import Dict, Iterable, List, Optional, Tuple, TypeVar, Union, overload
+from typing import Any, Dict, Iterable, List, Optional, Tuple, TypeVar, Union, overload
 from zipfile import ZipFile
 
 from frictionless import Detector, Layout, Package, Resource
-from glom import Assign, Coalesce, glom, Invoke, Iter, Spec, T
+from glom import Assign, Coalesce, glom, Invoke, Iter, Spec, SKIP, T
 from glom import Match, MatchError, Optional as optmatch, Or
 import pandas as pd
 
 from friendly_data.io import dwim_file, path_not_in, posixpathstr, relpaths
 from friendly_data.helpers import match, noop_map, is_windows
-from friendly_data._types import _path_t
+from friendly_data._types import _path_t, _dfseries_t
 import friendly_data_registry as registry
 
 logger = getLogger(__name__)
@@ -54,7 +53,7 @@ def fullpath(resource: Resource) -> Path:
     return Path(resource.basepath) / resource["path"]
 
 
-def _resource(spec: Dict, basepath: _path_t = "", infer=True) -> Resource:
+def resource_(spec: Dict, basepath: _path_t = "", infer=True) -> Resource:
     """Create a Resource object based on the dictionary
 
     Parameters
@@ -81,12 +80,14 @@ def _resource(spec: Dict, basepath: _path_t = "", infer=True) -> Resource:
         return spec
     assert "path" in spec, f"Incomplete resource spec: {spec}"
     opts = {}
+    layout_opts: Dict[str, Any] = {"header": True}
     if spec.get("skip"):
         # FIXME: `offset_rows` doesn't seem to work, so workaround with
         # `skip_rows` (`frictionless` expects a 1-indexed array).  `pandas` on
         # the other hand uses a 0-indexed list, which has to be accounted for
         # in `to_df`
-        opts["layout"] = Layout(skip_rows=[i + 1 for i in range(spec["skip"])])
+        layout_opts["skip_rows"] = [i + 1 for i in range(spec["skip"])]
+    opts["layout"] = Layout(**layout_opts)
     if spec.get("schema"):
         opts["detector"] = Detector(schema_patch=spec["schema"])
     res = Resource(path=str(spec["path"]), basepath=str(basepath), **opts)
@@ -106,8 +107,8 @@ def create_pkg(
 ):
     """Create a datapackage from metadata and resources.
 
-    If `resources` point to files that exist, their schema are inferred and
-    added to the package.  If `basepath` is a non empty string, it is treated
+    If ``resources`` point to files that exist, their schema are inferred and
+    added to the package.  If ``basepath`` is a non empty string, it is treated
     as the parent directory, and all resource file paths are checked relative
     to it.
 
@@ -118,7 +119,7 @@ def create_pkg(
 
     fpaths : Iterable[Union[str, Path, Dict]]
         An iterator over different resources.  Resources are paths to files,
-        relative to `basepath`.
+        relative to ``basepath``.
 
     basepath : str (default: empty string)
         Directory where the package files are located
@@ -154,7 +155,7 @@ def create_pkg(
         if not keep(spec["path"]):
             continue
         # NOTE: noop when Resource
-        _res = _resource(spec, basepath=basepath, infer=infer)
+        _res = resource_(spec, basepath=basepath, infer=infer)
         pkg.add_resource(_res)
 
     return _ensure_posix(pkg)
@@ -163,15 +164,15 @@ def create_pkg(
 def read_pkg(pkg_path: _path_t, extract_dir: Optional[_path_t] = None):
     """Read a  datapackage
 
-    If `pkg_path` points to a `datapackage.json` file, read it as is.  If it
+    If ``pkg_path`` points to a ``datapackage.json`` file, read it as is.  If it
     points to a zip archive.  The archive is first extracted before opening it.
-    If `extract_dir` is not provided, the current directory of the zip archive
-    is used.
+    If ``extract_dir`` is not provided, the current directory of the zip archive
+    is used.  If it is a directory, look for a ``datapackage.json`` inside.
 
     Parameters
     ----------
     pkg_path : Union[str, Path]
-        Path to the `datapackage.json` file, or a zip archive
+        Path to the ``datapackage.json`` file, or a zip archive
 
     extract_dir : Union[str, Path]
         Path to which the zip archive is extracted
@@ -183,14 +184,15 @@ def read_pkg(pkg_path: _path_t, extract_dir: Optional[_path_t] = None):
     Raises
     ------
     ValueError
-        When an unsupported file format (not JSON or ZIP) is provided
+        When an unsupported format (not a directory, JSON, or ZIP) is provided
+    FileNotFoundError
+        When a datapackage.json file cannot be found
 
     """
     pkg_path = Path(pkg_path)
     if pkg_path.suffix == ".json":
-        with open(pkg_path) as pkg_json:
-            basepath = f"{Path(pkg_path).parent}"
-            pkg = Package(json.load(pkg_json), basepath=basepath)
+        pkg_json = pkg_path
+        basepath = Path(pkg_path).parent
     elif pkg_path.suffix == ".zip":
         if extract_dir is None:
             extract_dir = pkg_path.parent
@@ -198,75 +200,25 @@ def read_pkg(pkg_path: _path_t, extract_dir: Optional[_path_t] = None):
             extract_dir = Path(extract_dir)
         with ZipFile(pkg_path) as pkg_zip:
             pkg_zip.extractall(path=extract_dir)
-            with open(extract_dir / "datapackage.json") as pkg_json:
-                pkg = Package(json.load(pkg_json), basepath=f"{extract_dir}")
+            pkg_json = extract_dir / "datapackage.json"
+            basepath = extract_dir
     elif pkg_path.is_dir():
-        with open(pkg_path / "datapackage.json") as pkg_json:
-            pkg = Package(json.load(pkg_json), basepath=str(pkg_path))
+        pkg_json = pkg_path / "datapackage.json"
+        basepath = pkg_path
+
     else:
-        raise ValueError(f"{pkg_path}: expecting a JSON or ZIP file")
+        raise ValueError(f"{pkg_path}: not a directory, JSON, or ZIP file")
+
+    if pkg_json.exists():
+        pkg = Package(dwim_file(pkg_json), basepath=f"{basepath}")
+    else:
+        msg = f"{pkg_json}: not found"
+        logger.error(msg)
+        raise FileNotFoundError(msg)
+
     return _ensure_posix(pkg)
 
 
-def update_pkg(pkg: Package, resource: str, schema_update: Dict, fields: bool = True):
-    """Update package resource schema
-
-    .. deprecated:: 0.2
-
-       This function is not used any more, instead the schema is updated per
-       resource in :func:`~friendly_data.dpkg.res_from_entry`.
-
-    Parameters
-    ----------
-    pkg : Package
-        Package object
-
-    resource : str
-        Resource name FIXME: cannot handle duplicate names in subdirectories
-
-    schema_update : Dict
-        Updated fields in the schema, if `field` is `False`, can be used to
-        update `missingValues`, or `primaryKey`.  When updating the schema, it
-        looks like this ('foo'/'bar' are names of the fields being updated)::
-
-          {
-              "foo": {
-                  "name": "foo",
-                  "type": "datetime",
-                  "format": "default"
-              },
-              "bar": {
-                  "name": "bar",
-                  "type": "integer",
-                  "format": "default"
-              }
-          }
-
-    fields : bool (default: True)
-        If the schema update is a field, or not
-
-    Returns
-    -------
-    bool
-        Return the `Package.valid` flag; `True` if the update was valid.
-
-    """
-    logger.warning("update_pkg: this function has been deprecated")
-    assert "resources" in pkg, "Package should have at least one resource"
-    res, *_ = [res for res in pkg["resources"] if res["name"] == resource]
-    if fields:
-        for field in res["schema"]["fields"]:
-            if field["name"] in schema_update:
-                field.update(schema_update[field["name"]])
-    else:
-        # FIXME: do the following checks w/o asserts
-        assert "fields" not in schema_update, "cannot add fields to schema"
-        # prevents from adding additional keys
-        assert set(schema_update) - {"primaryKey", "missingValues"} == set()
-        res["schema"].update(schema_update)
-    return pkg.metadata_valid
-
-
 class pkgindex(List[Dict]):
     """Data package index (a subclass of ``list``)
 
@@ -282,7 +234,7 @@ class pkgindex(List[Dict]):
     - "sheet": sheet name or position (0-indexed) to use as dataset (optional,
       Excel only)
 
-    While iterating over an index, always use :method:`~pkgindex.records` to
+    While iterating over an index, always use :meth:`~pkgindex.records` to
     ensure all necessary keys are present.
 
     """
@@ -296,8 +248,9 @@ class pkgindex(List[Dict]):
         "alias": {str: str},
         "sheet": Or(int, str),
         "iamc": str,  # FIXME: Regex("^[0-9a-zA-Z_ |-{}]+$"),
+        "agg": {str: [{"values": [str], "variable": str}]},
     }
-    _optional = ["idxcols", "name", "skip", "alias", "sheet", "iamc"]
+    _optional = ["idxcols", "name", "skip", "alias", "sheet", "iamc", "agg"]
 
     @classmethod
     def from_file(cls, fpath: _path_t) -> "pkgindex":
@@ -338,8 +291,8 @@ class pkgindex(List[Dict]):
         try:
             return glom(idx, Iter(record_match).all())
         except MatchError as err:
-            logger.exception(f"{err.args[1]}: bad key in index file")
-            raise
+            logger.error(f"{err.args[1]}: bad key in index file")
+            raise err from None
 
     @overload
     def _validate_keys(cls, keys: str) -> str:
@@ -373,16 +326,16 @@ class pkgindex(List[Dict]):
         Raises
         ------
         glom.MatchError
-            If `keys` has an unsupported value
+            If ``keys`` has an unsupported value
 
         """
         spec = {k: Coalesce(k, default=None) for k in self._validate_keys(keys)}
         return glom(self, Iter().map(spec).all())
 
     def get(self, key: str) -> List:
-        """Get the value of `key` from all records as a list.
+        """Get the value of ``key`` from all records as a list.
 
-        If `key` is absent, the corresponding value is set to ``None``.
+        If ``key`` is absent, the corresponding value is set to ``None``.
 
         Parameters
         ----------
@@ -392,7 +345,7 @@ class pkgindex(List[Dict]):
         Returns
         -------
         List
-            List of records with values corresponding to `key`.
+            List of records with values corresponding to ``key``.
 
         """
         return glom(self, [Coalesce(self._validate_keys(key), default=None)])
@@ -418,8 +371,7 @@ def get_aliased_cols(cols: Iterable[str], col_t: str, alias: Dict[str, str]) ->
     -------
     Dict
         Schema for each column, the column name is the key, and the schema is
-        the value; see the doctring of :func:`friendly_data.dpkg.index_levels`
-        for more.
+        the value; see the doctring of :func:`index_levels` for more.
 
     """
     alias = noop_map(alias if isinstance(alias, dict) else {})
@@ -430,17 +382,16 @@ def get_aliased_cols(cols: Iterable[str], col_t: str, alias: Dict[str, str]) ->
     return coldict
 
 
-_file_or_df_t = Union[_path_t, pd.DataFrame]
-
-
 def index_levels(
-    file_or_df: _file_or_df_t, idxcols: Iterable[str], alias: Dict[str, str] = {}
-) -> Tuple[_file_or_df_t, Dict]:
+    file_or_df: Union[_path_t, _dfseries_t],
+    idxcols: Iterable[str],
+    alias: Dict[str, str] = {},
+) -> Tuple[_dfseries_t, Dict]:
     """Read a dataset and determine the index levels
 
     Parameters
     ----------
-    file_or_df : Union[str, Path, pd.DataFrame]
+    file_or_df : Union[str, Path, pd.DataFrame, pd.Series]
         A dataframe, or the path to a CSV file
 
     idxcols : Iterable[str]
@@ -451,10 +402,11 @@ def index_levels(
 
     Returns
     -------
-    Tuple[Union[str, Path, pd.DataFrame], Dict]
+    Tuple[Union[pd.DataFrame, pd.Series], Dict]
 
-        Tuple of path to the dataset, and the schema of each column as a dictionary.
-        If `idxcols` was ["foo", "bar"], the dictionary might look like::
+        Tuple of the dataset, and the schema of each index column as a
+        dictionary.  If ``idxcols`` was ["foo", "bar"], the dictionary might look
+        like::
 
           {
               "foo": {
@@ -476,14 +428,37 @@ def index_levels(
 
     """
     coldict = get_aliased_cols(idxcols, "idxcols", alias)
+    # 4 cases:
+    # 1) in registry, and enum constraints set:
+    #    {"constraints": {"enum": And(list, len)}, str: str}
+    # 2) in registry, but enum constraints unset: `enum_partial` below
+    # 3) in registry, but enum not applicable (e.g. datetime): nothing to match
+    # 4) not in registry; this has a few possibilities, the type could be
+    #    string, integer, or boolean, but the metadata isn't readily available.
+    #    It could be deduced from the dataframe, but that is a bit messy as it
+    #    requires reverse lookup of `converters._pd_types`.  There should be a
+    #    cost to not using the registry, and this is it - it will not be enum
+    #    constrained:
+    #    And(
+    #        {"type": Or("string", "integer", "boolean"), str: str},
+    #        lambda i: "constraints" not in i,
+    #    )
+    #
     # select columns with an enum constraint where the enum values are empty
-    select_cols = match({"constraints": {"enum": []}, str: str})
+    enum_partial = {"constraints": {"enum": []}, str: str}
+    select_cols = match(enum_partial)
     cols = glom(coldict.values(), Iter().filter(select_cols).map("name").all())
-    if isinstance(file_or_df, pd.DataFrame):
+
+    if isinstance(file_or_df, (pd.DataFrame, pd.Series)):
+        if not cols:
+            return file_or_df, coldict
         diff = list(set(idxcols) - set(cols))
         idx = file_or_df.index.droplevel(diff)
     else:
-        idx = pd.read_csv(file_or_df, index_col=cols).index
+        file_or_df = pd.read_csv(file_or_df, index_col=cols)
+        if not cols:
+            return file_or_df, coldict
+        idx = file_or_df.index
     if isinstance(idx, pd.MultiIndex):
         levels = {col: list(lvls) for col, lvls in zip(idx.names, idx.levels)}
     else:
@@ -496,11 +471,36 @@ def index_levels(
     return file_or_df, coldict
 
 
+def set_idxcols(fpath: _path_t, basepath: _path_t = "") -> Resource:
+    """Create a resource object for a file, with index columns set
+
+    Parameters
+    ----------
+    fpath : Union[str, Path]
+        Path to a dataset (resource), e.g. a CSV file, relative to `basepath`
+
+    basepath : Union[str, Path], default: empty string (current directory)
+        Path to directory to consider as the data package basepath
+
+    Returns
+    -------
+    Resource
+        Resource object with the index columns set according to the registry
+
+    """
+    res = resource_({"path": fpath}, basepath=basepath)
+    all_idxcols = glom(registry.getall()["idxcols"], ["name"])
+    idxcols = glom(
+        res, ("schema.fields", Iter("name").filter(lambda i: i in all_idxcols).all())
+    )
+    return glom(res, Assign("schema.primaryKey", idxcols))
+
+
 def res_from_entry(entry: Dict, pkg_dir: _path_t) -> Resource:
     """Create a resource from an index entry.
 
     Entry must have the keys: ``path``, ``idxcols``, ``alias``; so use
-    :method:`pkgindex.records` to iterate over the index.
+    :meth:`pkgindex.records` to iterate over the index.
 
     Parameters
     ----------
@@ -538,10 +538,10 @@ def res_from_entry(entry: Dict, pkg_dir: _path_t) -> Resource:
         # FIXME: too broad; most likely this fails because of bad options
         # (e.g. index file entry), validate options and narrow scope of except
         logger.exception(f"error reading {entry}")
-        raise
+        raise err from None
     entry.update(schema={"fields": idxcoldict, "primaryKey": entry["idxcols"]})
     # FIXME: should we wrap this in a similar try: ... except: ...
-    res = _resource(entry, basepath=f"{pkg_dir}", infer=True)
+    res = resource_(entry, basepath=f"{pkg_dir}", infer=True)
     # set of value columns
     cols = glom(res.schema.fields, (Iter("name"), set)) - set(entry["idxcols"])
     coldict = get_aliased_cols(cols, "cols", entry["alias"])
@@ -555,6 +555,43 @@ def res_from_entry(entry: Dict, pkg_dir: _path_t) -> Resource:
     return res
 
 
+def entry_from_res(res: Resource) -> Dict:
+    """Create an index entry from a ``Resource`` object
+
+    Parameters
+    ----------
+    res : Resource
+        A resource object
+
+    Returns
+    -------
+    Dict
+        A dictionary that is an index entry
+
+    """
+    entry = glom(
+        res,
+        {
+            "name": "name",
+            "path": "path",
+            "idxcols": Coalesce("schema.primaryKey", default=SKIP),
+        },
+    )
+    alias = glom(
+        res,
+        (
+            "schema.fields",
+            Iter(match({"alias": str, str: object}))
+            .map(lambda i: (i["name"], i["alias"]))
+            .all(),
+            dict,
+        ),
+    )
+    if alias:
+        entry["alias"] = alias
+    return entry
+
+
 def pkg_from_index(meta: Dict, fpath: _path_t) -> Tuple[Path, Package, pkgindex]:
     """Read an index file, and create a datapackage with the provided metadata.
 
@@ -573,8 +610,8 @@ def pkg_from_index(meta: Dict, fpath: _path_t) -> Tuple[Path, Package, pkgindex]
 
     Returns
     -------
-    Tuple[Path, Package, pandas.DataFrame]
-        The package directory, the `Package` object, and the index.
+    Tuple[Path, Package, pkgindex]
+        The package directory, the ``Package`` object, and the index.
 
     Examples
     --------
@@ -615,7 +652,7 @@ def pkg_from_files(
     fpath : Union[str, Path]
         Path to the package directory or index file.  Note the index file has
         to be at the top level directory of the datapackage.  See
-        :func:`~friendly_data.dpkg.pkgindex.from_file`
+        :meth:`pkgindex.from_file`
 
     fpaths : List[Union[str, Path]]
         A list of paths to datasets/resources not in the index.  If any of the
@@ -700,7 +737,7 @@ def write_pkg(
         Path to write to
 
     idx : Union[pkgindex, List] (optional)
-        Package index written to `pkgdir/index.json`
+        Package index written to ``pkgdir/index.json``
 
     Returns
     -------
